using Microsoft.AspNetCore.Html;
using Microsoft.AspNetCore.Mvc.Rendering;
using Microsoft.AspNetCore.Mvc.Routing;
using Microsoft.CodeAnalysis.CSharp.Syntax;
using PhoneToolMX.Models;
using System.Linq.Expressions;
using Microsoft.AspNetCore.Mvc;

namespace PhoneToolMX.Helpers
{
    public static class HtmlTable
    {
        /// <summary>
        /// Creates an HTML table for a given collection of model objects.
        /// </summary>
        /// <param name="helper">HtmlHelper instance in the view</param>
        /// <param name="headers">List of table headers</param>
        /// <param name="rowFunc">Function used to determine the structure of each row. All objects will be converted to strings with ToString().</param>
        /// <typeparam name="TModel"></typeparam>
        /// <returns>An HTML table representing the list.</returns>
        /// <remarks>
        /// https://xkcd.com/1319/
        /// </remarks>
        public static IHtmlContent MakeTable<TModel>(this IHtmlHelper<ICollection<TModel>> helper)
        {
            var builder = new HtmlContentBuilder();
            var urlFactory = helper.ViewContext.HttpContext.RequestServices.GetRequiredService<IUrlHelperFactory>();
            var url = urlFactory.GetUrlHelper(helper.ViewContext);
            var props = typeof(TModel).GetProperties().Where(pi => pi.GetCustomAttributes(typeof(HeaderAttribute), false).Length != 0).ToList();
            var idProp = typeof(TModel).GetProperty("Id");
            var headerList = props.Select(pi =>
            {
                var attr = (HeaderAttribute)pi.GetCustomAttributes(typeof(HeaderAttribute), false)[0];
                return attr;
            }).ToList();
            headerList.Add(new HeaderAttribute
            {
                Title = "Actions",
                Small = true,
            });

            builder.AppendHtmlLine("<table id=\"listview\">")
                .AppendHtmlLine("<thead><tr>");
            foreach (var header in headerList) {
                builder.AppendFormat("<th{0}>{1}</th>",
                    (header.Small) ? " width=2" : string.Empty,
                    header.Title);
            }
            builder.AppendHtmlLine("</tr></thead>")
                .AppendHtmlLine("<tbody>");
            
            // build body
            if (helper.ViewData.Model is {} rows) {
                foreach (var row in rows) {
                    var cells = props
                        .Select(pi =>
                        {
                            var val = pi.GetValue(row, null);
                            if (val is bool someBool) {
                                // idk, yes/no feels better to me than true/false for user-facing stuff
                                return someBool ? "Yes" : "No";
                            }
                            return val?.ToString() ?? "(null)";
                        })
                        .ToList();
                    var cellId = idProp?.GetValue(row, null)?.ToString() ?? "INVALID_ID";
                    builder.AppendHtmlLine("<tr>");
                    foreach (var cell in cells) {
                        builder.AppendFormat("<td>{0}</td>", cell);
                    }
                    // Actions cell
                    builder.AppendHtml("<td align=\"right\">")
                        // ReSharper disable Mvc.ActionNotResolved
                        .AppendHtml($"<a href=\"{url.Action("Edit", new {id = cellId})}\"><img src=\"{url.Content("~/images/edit_x16.gif")}\" alt=\"Edit\" /></a>")
                        .AppendHtml($"<a href=\"{url.Action("Delete", new {id = cellId})}\"><img src=\"{url.Content("~/images/del_x16.gif")}\" alt=\"Delete\" /></a>")
                        .AppendHtmlLine("</td>")
                        .AppendHtmlLine("</tr>");
                }
                if (rows.Count == 0) {
                    builder.AppendFormat("<tr><td colspan=\"{0}\" align=\"center\">No entries found</td></tr>", headerList.Count);
                }
            } else {
                builder.AppendFormat("<tr><td colspan=\"{0}\" align=\"center\">Invalid model found</td></tr>", headerList.Count);
            }
            
            // close remaining table tags
            return builder.AppendHtmlLine("</tbody>")
                .AppendHtmlLine("</table>");
        }
    }
}
